// DTMNT emu - INPUT module

#include <stdio.h>
#include <stdlib.h>
#include <allegro.h>
#include "dtmnt.h"

int set_joy_type=JOY_TYPE_AUTODETECT;

// In order system, leo, mike, don, raph

int input_type[5][8]={
  {KEY_F1,KEY_F2,KEY_F3,KEY_F4,0,0,0,0},
  {0x100+KEY_LEFT,0x100+KEY_RIGHT,0x100+KEY_UP,0x100+KEY_DOWN,KEY_ALT,KEY_CONTROL,KEY_SPACE,KEY_F5},
  {0x1002,0x1003,0x1000,0x1001,0x1011,0x1010,0x1012,KEY_F6},
  {0x2002,0x2003,0x2000,0x2001,0x2011,0x2010,0x2012,0x2013},
  {0x3006,0x3007,0x3004,0x3005,0x3000,0x3008,0x3001,0x3003}
};

// SNESPAD: The order is B,Y,sel,start,up,down,left,right,A,X,L,R
//                       0 1 2   3     4  5    6    7     8 9 a b

// I've decided to code EVERYTHING (e.g. keys, joystick/pads, mouse)
// into once code (e.g. code<0x080 == keyboard scancode)

unsigned char input[5][8]={{0}};

static FILE *save_file=NULL,*load_file=NULL;
static int input_start=1;

int save_input_init(char *name)
{
  static char id[64]="DTmnt Movie -- v" VER;

  if (save_file==NULL)
  {
    save_file=fopen(name,"wb");
    if (save_file==NULL) return 1;
    if (fwrite(id,1,64,save_file)!=64) return 1;
    fseek(save_file,0x40,SEEK_CUR);
    if (machine_save_to(save_file)!=0) return 1;
    input_start=1;
  }
  return 0;
}

int load_input_init(char *name)
{
  static char id[64]="";

  if (load_file==NULL)
  {
    load_file=fopen(name,"rb");
    if (load_file==NULL) return 1;
    if (fread(id,1,64,load_file)!=64) return 1;
    fseek(load_file,0x40,SEEK_CUR);
    if (machine_load_from(load_file)!=0) return 1;
    input_start=1;
  }
  return 0;
}


int saveload_input_exit()
{
  if (save_file!=NULL)
  {
    fclose(save_file);
    save_file=NULL;
  }
  if (load_file!=NULL)
  {
    fclose(load_file);
    load_file=NULL;
  }

  return 0;
}

int save_input()
{
  int i;
  if (save_file!=NULL)
  {
    for (i=0;i<5;i++)
    {
      if (fwrite(&aoo[i],1,sizeof(aoo[i]),save_file)!=sizeof(aoo[i]))
        return 1;
    }
  }
  return 0;
}

int load_input()
{
  int i=0;
  if (load_file!=NULL)
  {
    for (i=0;i<5;i++)
    {
      if (fread(&aoo[i],1,sizeof(aoo[i]),load_file)!=sizeof(aoo[i]))
        return 1;
    }
  }

  return 0;
}

// SNES PAD in LPT stuff START -----------
static int snespad_press[64]={0};
static int port_lpt1=0x378,port_lpt2=0x278;
int lpt_read=0;

static int pport_help(int port_no,int vala,int valb)
{
  int result;
  outportb(port_no,vala);
  outportb(port_no,0xf8);
  result=inportb(port_no+1);
  result&=valb;
  return (result==0);
}

static int poll_snespad()
{
  int i=0;
  int buttons=12;

  if (lpt_read&1)
  {
    snespad_press[0]=  pport_help(port_lpt1,0xfa,0x40);
    for (i=1;i<buttons;i++)
      snespad_press[i]=pport_help(port_lpt1,0xf9,0x40);
  
    snespad_press[16]= pport_help(port_lpt1,0xfa,0x20);
    for (i=(16+1);i<(16+buttons);i++)
      snespad_press[i]=pport_help(port_lpt1,0xf9,0x20);
  }

  if (lpt_read&2)
  {
    snespad_press[32]=  pport_help(port_lpt2,0xfa,0x40);
    for (i=(32+1);i<(32+buttons);i++)
      snespad_press[i]=pport_help(port_lpt2,0xf9,0x40);
  
    snespad_press[48]= pport_help(port_lpt2,0xfa,0x20);
    for (i=(48+1);i<(48+buttons);i++)
      snespad_press[i]=pport_help(port_lpt2,0xf9,0x20);
  }

  return 0;
}

// SNES PAD in LPT stuff END -----------



static int dx=0,dy=0,m_dir=4;

int input_update_mouse()
{
  int ddx=0,ddy=0;
  int rx=0,ry=0;
  get_mouse_mickeys(&ddx,&ddy);

  dx=(dx*3)/4;
  dy=(dy*3)/4;

  dx+=ddx;
  dy+=ddy;

  rx=(dx<0) ? (-dx) : dx;
  ry=(dy<0) ? (-dy) : dy;

  // If the angle of motion is less than about 30 degrees, consider
  // it NOT diagonal
  if (rx>(ry*2)) dy=0;
  else if (ry>(rx*2)) dx=0;

  if (dx<0) m_dir=3;
  else if (dx==0) m_dir=4;
  else m_dir=5;

  if (dy<0) m_dir-=3;
  else if (dy==0) ;
  else m_dir+=3;

  return 0;
}

static int input_device(int code)
{
  if (code<=0) return 0;
  if (code<0x100) return (key[code&127]&1) ? 1 : 0;
  if (code<0x200) return (key[code&127]&2) ? 1 : 0;
  if ((code&0xff00)==0x3000) return snespad_press[code&63];

  switch (code)
  {
    case 0x1000: return joy_up;
    case 0x1001: return joy_down;
    case 0x1002: return joy_left;
    case 0x1003: return joy_right;
    case 0x1010: return joy_b1;
    case 0x1011: return joy_b2;
    case 0x1012: return joy_b3;
    case 0x1013: return joy_b4;
    case 0x1100: return joy2_up;
    case 0x1101: return joy2_down;
    case 0x1102: return joy2_left;
    case 0x1103: return joy2_right;
    case 0x1110: return joy2_b1;
    case 0x1111: return joy2_b2;
    case 0x2010: return ((mouse_b&1)?1:0);
    case 0x2011: return ((mouse_b&2)?1:0);
    case 0x2012: return ((mouse_b&4)?1:0);
    // mouse direction

    case 0x2000:
      if ((m_dir==0)||(m_dir==1)||(m_dir==2))
        return 1; else return 0;
    case 0x2001:
      if ((m_dir==6)||(m_dir==7)||(m_dir==8))
        return 1; else return 0;
    case 0x2002:
      if ((m_dir==0)||(m_dir==3)||(m_dir==6))
        return 1; else return 0;
    case 0x2003:
      if ((m_dir==2)||(m_dir==5)||(m_dir==8))
        return 1; else return 0;
    default:
  }
  return 0;
}

int w=0;

static int what_is_input(int tmnt_key)
{
  // waits for the user to press a key, joystick direction etc...
  // then returns the code for it
  int done=0,wait_for_mouseb_zero=0,code=0;
  while (!done)
  {
    int mb,i;

    // Don't want keyboard buffer to fill up
    while (keypressed()) (void)readkey();

    // check normal keys
    for (i=0;i<128;i++)
    {
      int status=key[i];
      if (status)
      {
        done=1;
        code=(i+ ((status&3)-1)*0x100); break;
      }
    }

    mb=mouse_b;
    if (mb)
    {
      if (tmnt_key<4) { code=0x2000+(tmnt_key^2); wait_for_mouseb_zero=1; }
        // mouse directions are special
      else
      {
        if (mouse_b&1) code=0x2010;  // buttons are normal
        if (mouse_b&2) code=0x2011;
        if (mouse_b&4) code=0x2012;
      }
      done=1;
    }
    poll_joystick();
    if (joy_up)    { code=0x1000; done=1; }
    if (joy_down)  { code=0x1001; done=1; }
    if (joy_left)  { code=0x1002; done=1; }
    if (joy_right) { code=0x1003; done=1; }
    if (joy_b1)    { code=0x1010; done=1; }
    if (joy_b2)    { code=0x1011; done=1; }
    if (joy_b3)    { code=0x1012; done=1; }
    if (joy_b4)    { code=0x1013; done=1; }
    if (joy2_up)   { code=0x1100; done=1; }
    if (joy2_down) { code=0x1101; done=1; }
    if (joy2_left) { code=0x1102; done=1; }
    if (joy2_right){ code=0x1103; done=1; }
    if (joy2_b1)   { code=0x1110; done=1; }
    if (joy2_b2)   { code=0x1111; done=1; }
    poll_snespad();
    for (i=0;i<64;i++)
    {
      if (snespad_press[i]) { code=0x3000+i; done=1; }
    }
  }

  if (code==KEY_ESC)
    printf (".");
  else
    printf (".%4x",code);

  fflush(stdout);

  // Wait a tenth of a sec:
  rest(100);

  // wait till it gets released
  if (wait_for_mouseb_zero)
  {
    while (mouse_b)
      ;
  }
  else
  {
    poll_joystick();
    poll_snespad();
    while (input_device(code))
    {
      if (key[KEY_ESC])
      {
        while (key[KEY_ESC])
          ;
        return KEY_ESC;
      }
      poll_joystick();
      poll_snespad();
    }
  }
  printf (".");
  fflush(stdout);
  return code;
}

static int redefine_input_for_player(int p)
{
  int k=0;
  int doto=6;
  char key_names[2][8][32]=
  {
    {"left","right","up","down","jump","attack","button 3","credit"},
    {"Coin1","Coin2","Coin3","Coin4",
     "Service1","Service2","Service3","Service4"
    }
  };

  doto=8;
  printf ("Redefining input for player %d\n",p);
  printf ("(Hit ESCAPE to skip and not change)\n\n");
  for (k=0;k<doto;k++)
  {
    int code=0;
    int rk=k;
    if (p>0)
    {
      // get a nicer order
      if (rk<4) rk^=2;
      else if (rk<6) rk^=1;
    }
    printf ("  %-8.8s (%4x): ",key_names[(p>0)?0:1][rk],input_type[p][rk]);

    fflush(stdout);
    code=what_is_input(rk);
    if (code==KEY_ESC)
      printf ("(skip)");
    else
      input_type[p][rk]=code;
    printf ("\n");
  }

  return 0;
}

int redefine_input()
{
  int gotkey=0,done=0;
  int old_pause=pause;
  pause=1;
  set_gfx_mode(GFX_TEXT,0,0,0,0);
  // Wait until waveout_stream needs more, then blank it (because we are paused)
  while (waveout_stream_update()==1)   ;
  // Blank the other half as well
  while (waveout_stream_update()==1)   ;

  printf ("DTMNT v%s\n\nRedefine keys:\n-------------\n\n",VER);

  done=0;
  while (!done)
  {
    int player=-1;
    printf (
    "\nWhich player 1-4? (0 for coins) ENTER to resume emulation.\n");

    while (!keypressed())
      ;
    while (keypressed())
      gotkey=readkey()>>8;

    if (gotkey==KEY_ENTER) done=1;
    else if (gotkey==KEY_ESC) done=1;
    else if (gotkey==KEY_1) player=1;
    else if (gotkey==KEY_2) player=2;
    else if (gotkey==KEY_3) player=3;
    else if (gotkey==KEY_4) player=4;
    else if (gotkey==KEY_0) player=0;

    // wait till key is released
    while (key[gotkey&127])
      ;

    if (player>=0)
      redefine_input_for_player(player);
  }

  set_color_depth(actual.depth);
  set_gfx_mode(actual.driver,actual.x,actual.y,0,0);
  if (actual.depth<=8) set_palette(game_pal);

  pause=old_pause;
  return 0;
}

int update_input()
{
  int p,k;
  unsigned char val=0; int i=0;
  poll_joystick();
  poll_snespad();
  // Freeze *this* moment and save current control status in input[5][8]
  for (p=0;p<5;p++)
  {
    for (k=0;k<8;k++)
    {
      input[p][k]=0;
      if (input_device(input_type[p][k]))
        input[p][k]=1;
    }
  }

  // precalc a0000 info
  val=0;
  for (i=0;i<8;i++)
    {val<<=1; val+=input[0][7-i] ? 1 : 0 ;}
  aoo[0]=val;

  val=0;
  for (i=0;i<8;i++)
    {val<<=1; val+=input[1][7-i] ? 1 : 0 ;}
  aoo[1]=~val;

  val=0;
  for (i=0;i<8;i++)
    {val<<=1; val+=input[2][7-i] ? 1 : 0 ;}
  aoo[2]=~val;

  val=0;
  for (i=0;i<8;i++)
    {val<<=1; val+=input[3][7-i] ? 1 : 0 ;}
  aoo[3]=~val;

  val=0;
  for (i=0;i<8;i++)
    {val<<=1; val+=input[4][7-i] ? 1 : 0 ;}
  aoo[4]=~val;

  return 0;
}
