// DTMNT emu - Sound emulation

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>
#include "fm.h"
#include "mz80.h"
#include "dtmnt.h"

int want_sound=1,do_sound=1;
int timerb_speed=250;
int waveout_size=1024; // In samples
int waveout_rate=16000;// In samples/sec
int stereo16bit=0;

// In sbsnd.c
extern int play_sbsnd_stream
  (int size,int bits,int stereo,int rate,int vol,int pan);
extern int stop_sbsnd_stream();
extern void *get_sbsnd_stream_buffer();
extern int free_sbsnd_stream_buffer();

int channels_on=0xff;
int ym2151_amp=19;
int drum_volume=26;

int using_digi_card=100; // daves sbsnd
int get_from_blaster=1;
AUDIOSTREAM *audio_stream=NULL;
int sound_effect_count=0;
unsigned char sound_effect[SOUND_EFFECT_LIMIT]={0};
int force_dsp_version=0;
extern int externally_force_dsp_version;

int dodgy_drums=1;
static unsigned char *drums=NULL;
static unsigned char *voice=NULL;
static unsigned char *sample_d05=NULL;
static unsigned char *fmbuf=NULL;
static int grab_count=8;
static unsigned char temp[1024]="";
static int adshift=18;
static int *mixing=NULL;

static int read_blaster_env()
{
  extern int sbsnd_base,irq_num,dma_chan_8,dma_chan_16;
  char *blast,*next;
  int done=0;
  blast=getenv("BLASTER");
  if (blast==NULL)
  {
    printf ("No BLASTER variable.\n");
    return 1;
  }

  printf ("Getting from BLASTER var:  ");

  done=0;
  while (!done)
  {
    next=blast+1;
    if (*blast==0) done=1;
    else
    {
      switch (*blast)
      {
        case 'a': case 'A':
          sbsnd_base=strtol(blast+1,&next,16);
          printf ("A:%x  ",sbsnd_base);
        break;
        case 'd': case 'D':
          dma_chan_8=strtol(blast+1,&next,0);
          printf ("D:%d  ",dma_chan_8);
        break;
        case 'h': case 'H':
          dma_chan_16=strtol(blast+1,&next,0);
          printf ("H:%d  ",dma_chan_16);
        break;
        case 'i': case 'I':
          irq_num=strtol(blast+1,&next,0);
          printf ("I:%d  ",irq_num);
        break;
      }
    }
    blast=next;
  }
  printf (".\n");

  return 0;
}


int sound_init()
{
  extern int rom_set;

  // stupidity protection!
  if (waveout_size<16) waveout_size=16;
  if (waveout_rate<16) waveout_rate=16;

  if (get_from_blaster) read_blaster_env();

  do_sound=want_sound;

  if (do_sound)
  {
    if (using_digi_card==100)
      printf ("Using Dave's SoundBlaster driver: SBSND\n");
    else
    {
      printf ("Using Allegro for sound: driver number is %d\n",using_digi_card);
      // Use allegro for sound, not SBSND
      if (force_dsp_version>0) // VERY IMPORTANT! The allegro hack had >=0
        externally_force_dsp_version=force_dsp_version;
      if (install_sound(using_digi_card,MIDI_NONE,NULL)!=0)
      {
        printf ("  Failed to install sound\n Allegro says: %s\n",allegro_error);
        do_sound=0;
      }
      else
      {
        if (digi_driver==NULL) printf (" Sound Driver is null(?)\n");
        else printf (" Sound Driver is %s\n",digi_driver->name);
      }
    }
  }

  if (do_sound)
  {
    int ret=0;
    printf ("Rendering is %s\n",stereo16bit?"stereo 16-bit":"mono 8-bit");
    if (using_digi_card==100)
    {
      // Use SBSND for sound
  
      if (stereo16bit)
        ret=play_sbsnd_stream(waveout_size,16,1,waveout_rate,255,128);
      else
        ret=play_sbsnd_stream(waveout_size,8,0,waveout_rate,255,128);
  
      if (ret!=0)
      {
        printf ("play_sbsnd_stream() failed.\n");
        do_sound=0;
      }
    }
    else
    {
      // Use Allegro for sound
      if (stereo16bit)
        audio_stream=play_audio_stream(waveout_size,16,1,waveout_rate,255,128);
      else
        audio_stream=play_audio_stream(waveout_size,8,0,waveout_rate,255,128);
      if (audio_stream==NULL)
        do_sound=0;
    }
  }

  if (do_sound)
  {
    // Allocate mixing buffer
    mixing=tralloc(waveout_size*sizeof(int));
    if (mixing==NULL)
    {
      printf ("Couldn't allocate mixing buffer.\n");
      do_sound=0;
    }
  }

  if (do_sound)
  {
    extern double alter_pitch;
    int ret=0;
    fmbuf=tralloc(waveout_size);

    // Grab is to split the waveform up so aliasing doesn't occur
    // Split into about one every 1/80 sec
    grab_count=(waveout_size*80)/waveout_rate;
    if (grab_count<1) grab_count=1;
    printf ("(grab_count is %d)\n",grab_count);

    // I think the buffer size is needed so that it knows how often
    // updateone will be called
// 3.579545 according to JesusLA - put it here too
    ret=OPMInit(1,(double)3579645*alter_pitch,waveout_rate,stereo16bit?16:8,waveout_size/grab_count,(void **)&fmbuf);
    printf ("OPMInit() returned %d\n",ret);
    OPMResetChip(0);
  }

  if (do_sound)
  {
    strcpy(temp,rom_path); 
    switch(rom_set)
    {
      case 1: case 2: case 3: case 4: strcat(temp,"963-a26"); break;
      default: strcat(temp,"tmnt.c13");
    }
    drums=tralloc(128*1024);
    if (drums==NULL)
    { printf ("out of memory\n"); do_sound=0; }
    else
    {
      memset(drums,128,128*1024);
      load_rom(drums,temp,1,128*1024);
    }
  }

  if (do_sound)
  {
         if (waveout_rate>=20000) adshift=18;
    else if (waveout_rate>=8000)  adshift=19;
    else                          adshift=20;

    strcpy(temp,rom_path);
    switch(rom_set)
    {
      case 1: case 2: case 3: case 4: strcat(temp,"963-a27"); break;
      default: strcat(temp,"tmnt.d18");
    }
    voice=tralloc(128*1024);
    if (voice==NULL)
    { printf ("out of memory\n"); do_sound=0; }
    else
    {
      memset(voice,0,128*1024);
      load_rom(voice,temp,1,128*1024);
    }
  }



  if (do_sound)
  {
    unsigned char *rom_d05=tralloc(0x80000);
    if (rom_d05==NULL)
    {
      printf ("out of memory - rom_d05\n");
    }
    else
    {
      strcpy(temp,rom_path);
      switch(rom_set)
      {
        case 1: case 2: case 3: case 4: strcat(temp,"963-a25"); break;
        default: strcat(temp,"tmnt.d05");
      }
      if (load_rom(rom_d05,temp,1,0x80000)!=0)
      {
        printf ("Error loading tmnt.d05\n");
      }
      else
      {
        sample_d05=tralloc(0x40000);
        if (sample_d05==NULL)
        {
          printf ("out of memory - sample_d05\n");
        }
        else
          decode_theme_music(rom_d05,0x80000,sample_d05,8);
      }
      tree(rom_d05);
    }
  }

  printf ("  do_sound is %s\n",do_sound ? "on":"off");
  return (do_sound==0);
}

int sound_exit()
{
  if (do_sound)
  {
    OPMShutdown();
    if (fmbuf!=NULL) tree(fmbuf);
    if (using_digi_card==100)
    {
      stop_sbsnd_stream();
    }
    else
      remove_sound();
    tree(mixing); mixing=NULL;
  }
  if (drums!=NULL) tree(drums);
  if (voice!=NULL) tree(voice);
  if (sample_d05!=NULL) tree(sample_d05);
  do_sound=0;

  return 0;
}

static int done_with_effect()
{
  // Shuffle down the rest of the sound effects
  int i;
  if (sound_effect_count<=0) return 1;
  for (i=(sound_effect_count-2); i>=0; i--)
  {
    sound_effect[i]=sound_effect[i+1];
  }
  sound_effect_count--;
  return 0;
}

static int adpcm_effect=-1;
int voicepos[64]=
{
  0,
  3845,
  6930,
  10681,
  14268,
  17163,
  19038,
  23467,
  40663,
  43459,
  48634,//aa
  55900,
  60600,
  64782,
  78063,
  89553,
  106886,
  118769,
  125200,
  131737,
  136320,
  144113,
  150482,
  154218,
  162523,
  165187,
  167960,
  172630,
  185818,
  193919,
  202073,
  213812,
  221438,
  236711,
  244488,
  255545,
  65536*4
};

int adpcm_sound(int *into,int len)
{
  int i,v;
  static struct {unsigned at,end,speed; int pos; } adchan[4]={{0}};
  int ad=0;
//0x1ffff
// Use 18.13 accuracy (sample position)
// Use 15.16 for the pos
  if (adpcm_effect>=0)
  {
    for (v=0;v<4;v++)
    {
      if (adchan[v].at<adchan[v].end) continue; // Being used
      adchan[v].pos=0;
      adchan[v].at=voicepos[(adpcm_effect-0xa0)&63]<<13;
      adchan[v].end=voicepos[(adpcm_effect-0xa0+1)&63]<<13;
      switch (adpcm_effect)
      {
        case 0xa0: case 0xa1: case 0xa2: case 0xa3:
        case 0xa4: case 0xa5: case 0xa6: case 0xa8:
          adchan[v].speed=(((long long)12000)<<13)/waveout_rate;
          break;
        case 0xb3:
          adchan[v].speed=(((long long)5500)<<13)/waveout_rate;
          break;
        case 0xb7:
          adchan[v].speed=(((long long)8000)<<13)/waveout_rate;
          break;
        case 0xba:
          adchan[v].speed=(((long long)7000)<<13)/waveout_rate;
          break;
        case 0xbb:
          adchan[v].speed=(((long long)8000)<<13)/waveout_rate;
          break;
        case 0xc2:
          adchan[v].speed=(((long long)5500)<<13)/waveout_rate;
          break;
        default:
          adchan[v].speed=(((long long)6000)<<13)/waveout_rate;
      }
      adpcm_effect=-1;
      break;
    }
  }

  for (v=0;v<4;v++)
  {
    for (i=0; i<len; i++)
    {
      int pos;
      if (adchan[v].at>=adchan[v].end) { adchan[v].at=0; adchan[v].end=0; break; }
      ad=voice[adchan[v].at>>14]; // two samples per byte
      // Split into two 4-bit ADPCM samples
      if ((adchan[v].at&0x2000)==0) ad>>=4;
      ad&=15;
      adchan[v].pos*=7; // fade a bit to keep in range
      adchan[v].pos>>=3;
      if (ad&8) adchan[v].pos+=((ad&7)<<adshift); else adchan[v].pos-=((ad&7)<<adshift);
      pos=adchan[v].pos;
      into[i]+=pos>>8;

      adchan[v].at+=adchan[v].speed;
    }
  }

  return 0;
}

static int play_theme=0;
static struct { unsigned at,end,speed,vol; } drum[2];

/*
static unsigned int drumpos[16]={
    0, 8278,//kick
 8394,17838,//snare
18071,22035,//hihat
22268,34393,//tom?
34743,46751,//chord/orch hit
46868,54096,//"hey"
54213,61091,//"wow"
61324,64006 //"check"?
};
*/

// Play a drum on channel 0 or 1
int play_a_drum(int drumaddr,int speed,int vol,int d)
{
  if (!dodgy_drums) return 1;
  if ((drumaddr<0)||(drumaddr>=0x20000)) return 1;

  d&=1;

  // Drums Use 20.12 accuracy
  drum[d].at=drumaddr<<12;

  {
    int j=0;
    // Find the drum's end (bit 7 set) beforehand
    for (j=drumaddr;j<0x20000;j++)
      if (drums[j]&0x80) break;
    drum[d].end=j<<12;
  }
  drum[d].speed=(((long long)speed)<<12)/waveout_rate;
  drum[d].vol=vol*drum_volume;
  if (drum[d].speed<=0) { drum[d].at=0; drum[d].end=0; }

  return 0;
}

int other_sound(int *into,int len)
{
  extern unsigned char *sample_d05;
//0x7fff8000 use 16.15 accuracy 
  static int theme_at=0,theme_go=0,theme_speed=0;
  int d;
  static struct sam16 *into16;
  into16=(struct sam16 *)into;

  if (play_theme)
  {
    if (play_theme>0)
    {
      theme_at=0; theme_go=1;
      theme_speed=(((long long)21000)<<12)/waveout_rate;
    }
    else theme_go=0;
    play_theme=0;
  }

// Use 19.12 accuracy
  if (theme_at>=(0x40000<<12)) theme_go=0;
  if (theme_go)
  {
    int i;
    for (i=0; i<len; i++)
    {
      int pos;
      if (theme_at>=(0x40000<<12)) break;
      pos=(sample_d05[(theme_at>>12)&0x3ffff]-128)>>2;
      into[i]+=(pos<<10);
      theme_at+=theme_speed;
    }
  }

  // Drums Use 20.12 accuracy

  for (d=0;d<2;d++)
  {
    unsigned drum_at,drum_end,drum_speed,drum_vol;
    if (drum[d].at>=drum[d].end) {drum[d].at=0; drum[d].end=0; continue;}
    if (drums==NULL) continue;

    drum_at=drum[d].at;
    drum_end=drum[d].end;
    drum_speed=drum[d].speed;
    drum_vol=drum[d].vol;

    asm(
  
  // int i;
  //      for (i=0; i<len; i++)
  //      {
      "\n  movl $0,%%ecx\n"
      "drum_loop:\n"
      "  cmpl %5,%%ecx\n"
      "  jge drum_finish\n"
  //        if (drum_at>=drum_end)
  //          break;

  
      "  movl %0,%%eax\n"
      "  cmpl %2,%%eax\n"
      "  jge drum_finish\n"
//  (((drums[(drum_at>>12)+0]-64)*vol)<<>>wahtever);
      "  xorl %%eax,%%eax\n"
      "  movl %4,%%esi\n"
      "  movl %0,%%edx\n"
      "  sarl $12,%%edx\n"
      "  movb 0(%%esi,%%edx,1),%%al\n"
      "  subl $64,%%eax\n"
      "  imull %7,%%eax\n"
//        into16[i]+=(above) left and right
      "  xorl %%ebx,%%ebx\n"
      "  movl %3,%%esi\n"
      "  movl 0(%%esi,%%ecx,4),%%ebx\n"
      "  addl %%eax,%%ebx\n"
      "  movl %%ebx,0(%%esi,%%ecx,4)\n"

  //        drum_at+=drum_speed;
      "  movl %6,%%ebx\n"
      "  addl %%ebx,%0\n"
     
  //      }
      "  addl $1,%%ecx\n"
      "  jmp drum_loop\n"
      "drum_finish:\n"
      : "=m" (drum_at)
      : "0" (drum_at) , "m" (drum_end), "m" (into), "m" (drums),
        "m" (len), "m" (drum_speed), "m" (drum_vol)
      : "%esi","%edx","%ecx","%ebx","%eax","%cc"
    );

    drum[d].at=drum_at;
    drum[d].end=drum_end;
    drum[d].speed=drum_speed;
    drum[d].vol=drum_vol;
  }// end of d=0;d<DRUM_COUNT loop

  return 0;
}


int try_to_pass_effect()
{
  if (sound_effect_count>0)
  {
    extern int z80_interrupt_pending;
    // We didn't handle the code with samples
    // Give code to the z80 - if it's music or synthed effects
    // YM2151 can handle it

    // Interrupt state is RIGHT but only gets turned on VERY rarely for
    // a split second...

    if (!code_waiting_for_z80)
    {
      sent_to_a0009=sound_effect[0];
      code_waiting_for_z80=1;
      z80_interrupt_pending=1;

      if (sent_to_a0009>=0xa0)
        if (sent_to_a0009<=0xc3)
          adpcm_effect=sent_to_a0009;

      if (sent_to_a0009==0x7d) play_theme=1;
      else if (sent_to_a0009==0x7e) play_theme=-1;

      done_with_effect();
    }
  }

  return 0;
}


static int mixing_to_more(int *mix,void *more,int len,int st16bit)
{
  int i,dis;
  // Take the int *mix sound buffer, and downgrade it to the format
  // Needed by more (i.e. 0-255 or 0-65535)
  if (st16bit)
  {
    unsigned short *into;
    into=(unsigned short *)more;
    for (i=0;i<len;i++)
    {
      dis=*(mix++); dis+=dis<<1; dis>>=2;
      dis+=0x8000;
      if (dis>0xffff) dis=0xffff;
      if (dis<0) dis=0;
      *(into++)=dis;
      *(into++)=dis;
    }
  }
  else
  {
    unsigned char *into;
    into=(unsigned char *)more;
    for (i=0;i<len;i++)
    {
      dis=*(mix++); dis+=dis<<1;
      dis>>=10; dis+=0x80;
      if (dis>0xff) dis=0xff;
      if (dis<0) dis=0;
      *(into++)=dis;
    }
  }

  return 0;
}

// Returns 1 if it is still playing last bit
int waveout_stream_update()
{
  static int z80_needs_to_do=0;
  int grab_no=0;
  int next_grab_at=0;
  static int next_timerb_at=0;
  int z80_has_done;
  int till_timerb=0;
  int till_grab=0;
  int till_z80_done=0;
  static int timera_frac=0;
  unsigned char *more=NULL;

  if (!do_sound) return 0;

  // didn't need any more yet
  if (using_digi_card==100)
    more=get_sbsnd_stream_buffer();
  else
    more=get_audio_stream_buffer(audio_stream);

  if (more==NULL) return 1;

  if (pause)
  {
    // **** I'd prefer to do this a faster way!
    if (mixing!=NULL)
    {
      memset(mixing,0,waveout_size*sizeof(int));
      mixing_to_more(mixing,more,waveout_size,stereo16bit);
    }
    if (using_digi_card==100)
      free_sbsnd_stream_buffer();
    else
      free_audio_stream_buffer(audio_stream);

    return 0;
  }

// Okay - we have the z80 going in the background
// It wants to do 'z80_needs_to_do' cycles before we free_audio
// The timer wants to interrupt it 200 times a second

// WE want to grab 'grab_count' bits of streamed sample from the YM2151
// while all this is going on...

// timer a about 50 times a second? < TMNT doesn't seems to notice it?
// timer b about 200 times a second?

  z80_needs_to_do+=((long long)SPEED_OF_Z80*virtual_fps*waveout_size)/waveout_rate/60;

  grab_no=0;

  z80_has_done=0;
  while (grab_no<grab_count)
  {
    int how_much_to_do=0,what_to_do_after=0;
    // How many cycles to do, and what to do after
    // 0 is nothing, 1 is timerb, 2 is sample grab

    // next time to grab a piece of sample
    next_grab_at=((long long)(grab_no+1)*z80_needs_to_do)/grab_count;

    till_timerb  =next_timerb_at -z80_has_done;
    till_grab    =next_grab_at   -z80_has_done;
    till_z80_done=z80_needs_to_do-z80_has_done;


    if (till_z80_done<till_grab)
    {
      // z80 finishes before grab wanted

      if (till_z80_done<till_timerb)
      {
        // Finish z80 off
        how_much_to_do=till_z80_done;
        what_to_do_after=0;
      }
      else
      {
        // do timerb after
        how_much_to_do=till_timerb;
        what_to_do_after=1;
      }
    }
    else
    {
      // grab wanted before or as z80 finishes
      if (till_grab<till_timerb)
      {
        how_much_to_do=till_grab;
        what_to_do_after=2;

      }
      else
      {
        // just do timerb after - grab waits till next loop
        how_much_to_do=till_timerb;
        what_to_do_after=1;
      }
    }

    if (how_much_to_do>30)
    {
      int ret=0,did=0;

      // Neil fixed the mz80 doing too many cycles problem
      try_to_pass_effect();

      ret = mz80exec(how_much_to_do);
      did += mz80GetElapsedTicks(TRUE);

      z80_has_done += did;

    }

    if (what_to_do_after==1)
    {
      // timerb needed (throw in timera too every fourth time)
      if (timera_frac==0) ym2151_timers|=1;
      timera_frac++; if (timera_frac>=4) timera_frac=0;
      ym2151_timers|=2;
      next_timerb_at+= (long long)SPEED_OF_Z80*virtual_fps/timerb_speed/60;
    }
    else if (what_to_do_after==2)
    {
      int st=0,en=0;
      st=((long long)grab_no*waveout_size)/grab_count;
      en=((long long)(grab_no+1)*waveout_size)/grab_count;
      OPMUpdateOne(0,mixing+st,en-st);
      other_sound(mixing+st,en-st);
      adpcm_sound(mixing+st,en-st);
      grab_no++;
    }
  }

  z80_needs_to_do-=z80_has_done;

  if (next_timerb_at>=z80_has_done) next_timerb_at-=z80_has_done;

  mixing_to_more(mixing,more,waveout_size,stereo16bit);

  if (using_digi_card==100)
    free_sbsnd_stream_buffer();
  else
    free_audio_stream_buffer(audio_stream);
  return 0;
}


int draw_waveout(BITMAP *bmp)
{
  extern unsigned char *sbsnd_buffer;
  extern int dmabuf_size;
  int offx,offy;
  int x,y,oldy;
  int white,black;
  int col;
  offx=bmp->w-256;
  offy=bmp->h-65;
  if (sbsnd_buffer==NULL) return 1;

  white=makecol_depth(bitmap_color_depth(bmp),255,255,255);
  black=makecol_depth(bitmap_color_depth(bmp),80,80,80);

  for (col=0;col<2;col++)
  {
    oldy=128;
    for (x=0;x<256;x++)
    {
      int ind;
      ind=(x*(dmabuf_size/2))>>8;
      if (stereo16bit)
      {
        y=sbsnd_buffer[(ind&-4)+1];
        y+=sbsnd_buffer[(ind&-4)+3];
        y>>=1;
      }
      else
        y=sbsnd_buffer[ind];
      // y is 0-255
      y>>=2;
      if (x>1)
        line(bmp,x-1+offx,oldy+offy+1-col,x+offx+1-col,y+offy+1-col,col?white:black);
      oldy=y;
    }
  }

  return 0;
}
