// DTMNT emu - graphics

#include <allegro.h>
#include <stdio.h>
#include <math.h>
#include "dtmnt.h"

// 64 sprites at 0x140400 onwards

//static unsigned char temp[1024];

struct vpage
{
  int offx,offy;
};

static int our_color[64*16]={0};
static unsigned long tmnts_last_color[64*16]={0};
static int our_last_depth[64*16]={0};

struct vpage vpage[4]={{0}};
// text 0x100000 == vpage[0]
// fg   0x102000 == vpage[1]
// bg   0x104000 == vpage[2]

// Deduce the scroll position for each vpage
static int update_vpage(int n)
{
  n&=3;
  vpage[n].offx=0;
  vpage[n].offy=0;

  if (n>0)
  {
    vpage[n].offx=MEMB(vidmem,0x006402+n-1);// 0xff 0x00 or 0x01
    vpage[n].offx<<=8;
    vpage[n].offx+=MEMB(vidmem,0x006400+n-1);

    if (vpage[n].offx>0x8000)
      vpage[n].offx-=0x10000;

    vpage[n].offy=MEMB(vidmem,0x006018+n-1);
  }
  vpage[n].offx+=12*8;
  vpage[n].offy+=2*8;
  
  return 0;
}

int gamma_correct=0;
double gamma_value=2.2;

// A TMNT color to our color codes
static inline int convert_to_our_color(unsigned long entry,int depth)
{
  int r,g,b;

  r = entry&0x1f;
  g =(entry&0x30000) >> 13;
  g|=(entry&0xe0) >> 5;
  b =(entry&0x7c0000) >> 18;

  // 5 bits of rgb now
  // any gamma correction needed?
  if (gamma_correct)
  {
    double fr=(r/32.0),fg=(g/32.0),fb=(b/32.0);
    double value=gamma_value;

    // frgb is 0.0 to <1.0
    fr=pow(fr,1.0/value);
    fg=pow(fg,1.0/value);
    fb=pow(fb,1.0/value);
    r=(fr*32.0); g=(fg*32.0); b=(fb*32.0);
    if (r<0) r=0;
    if (g<0) g=0;
    if (b<0) b=0;
    if (r>31) r=31;
    if (g>31) g=31;
    if (b>31) b=31;
  }

  r<<=3;g<<=3;b<<=3;
  return makecol_depth(depth,r,g,b);
}

// Convert ALL 64 palettes to our color values
static int convert_all_to_our_colors(int depth)
{
  int i;
  for (i=0;i<64*16;i++)
  {
    unsigned long col;
    col=MEML(colmem,0x00000+(i<<2));

    // Use a cache: only convert if this color is new
    // (conversion isn't as fast as it could be)
    if ((col!=tmnts_last_color[i])
       ||(depth!=our_last_depth[i]))
      our_color[i]=convert_to_our_color(col,depth);

    tmnts_last_color[i]=col;
    our_last_depth[i]=depth;
  }
  return 0;
}


static int tiledecoder1=0;
static int tiledecoder2=0;

#define REPEAT8(x) x x x x x x x x

// Draw the tile pointed to by b, with the palette (our style) pointed
// to by pal
static inline int draw_tile_slow(BITMAP *bmp,int tx,int ty,
  unsigned char *b,int *pal)
{
  int x,y;

  // Slow but sure method using putpixel
  unsigned long pi;
  for (y=ty;y<(ty+8);y++)
  {
    x=tx;
    // Get the next eight palette indexes
    pi=*((unsigned long *)b); b+=4;
    if (pi==0) continue;
#define SLOW_WAY \
   if (pi&15) putpixel(bmp,x,y,pal[pi&15]); x++; pi>>=4;

    REPEAT8(SLOW_WAY)
  }
  return 0;
}

// opaque version
#define ASM_CORE_8(PIX) \
"  movb %%bl,%%cl\n"\
"  andl $15,%%ecx\n"\
"  movb (%2,%%ecx,4),%%cl\n"\
"  movb %%cl,%%fs:(%3)\n"\
"  addl $1,%3\n"\
"  rorl $4,%%ebx\n\n"

static int draw_tile_8bpp(BITMAP *bmp,int tx,int ty,
  unsigned char *b,int *pal)
{
  unsigned long pi;
  unsigned char *dest;
  int add_to_get_to_next_line;

  add_to_get_to_next_line=(bmp->line[1]-bmp->line[0])-8;
  // -8 to subtract the 8 pixel we drew

  dest=(bmp->line[ty])+tx;

  REPEAT8(

    // Get the next eight palette indexes
    pi=*((unsigned long *)b); b+=4;
    asm(
      "\n  movw %w0,%%fs\n"
      ASM_CORE_8("0") ASM_CORE_8("1")
      ASM_CORE_8("2") ASM_CORE_8("3")
      ASM_CORE_8("4") ASM_CORE_8("5")
      ASM_CORE_8("6") ASM_CORE_8("7")
      :
      : "rm" (bmp->seg) , "b" (pi) , "r" (pal), "r" (dest)
      : "%ecx","%ebx","%cc"
    );
    dest+=add_to_get_to_next_line;
  )

  return 0;
}

// This is transparent
#define ASM_CORE_8_TRANS(PIX) \
"  movb %%bl,%%cl\n"\
"  andl $15,%%ecx\n"\
"  testb $15,%%ecx\n"\
"  jz trans_" PIX "\n"\
"  movb (%2,%%ecx,4),%%cl\n"\
"  movb %%cl,%%fs:(%3)\n"\
"  trans_" PIX ":\n"\
"  addl $1,%3\n"\
"  rorl $4,%%ebx\n\n"

static int draw_tile_8bpp_trans(BITMAP *bmp,int tx,int ty,
  unsigned char *b,int *pal)
{
  int y;
  unsigned long pi;
  unsigned char *dest;

  for (y=ty;y<(ty+8);y++)
  {
    dest=(bmp->line[y])+tx;

    // Get the next eight palette indexes
    pi=*((unsigned long *)b); b+=4;
    if (pi==0) continue;
    asm(
      "\n  movw %w0,%%fs\n"
      ASM_CORE_8_TRANS("0") ASM_CORE_8_TRANS("1")
      ASM_CORE_8_TRANS("2") ASM_CORE_8_TRANS("3")
      ASM_CORE_8_TRANS("4") ASM_CORE_8_TRANS("5")
      ASM_CORE_8_TRANS("6") ASM_CORE_8_TRANS("7")
      :
      : "rm" (bmp->seg) , "b" (pi) , "r" (pal), "r" (dest)
      : "%ecx","%ebx","%cc"
    );
  }
  return 0;
}

// opaque version
#define ASM_CORE_16(PIX) \
"  movb %%bl,%%cl\n"\
"  andl $15,%%ecx\n"\
"  movw (%2,%%ecx,4),%%cx\n"\
"  movw %%cx,%%fs:(%3)\n"\
"  addl $2,%3\n"\
"  rorl $4,%%ebx\n\n"

static int draw_tile_16bpp(BITMAP *bmp,int tx,int ty,
  unsigned char *b,int *pal)
{
  unsigned long pi;
  unsigned char *dest;
  int add_to_get_to_next_line;

  add_to_get_to_next_line=(bmp->line[1]-bmp->line[0])-16;
  // -16 to subtract the 8 pixel we drew

  dest=(bmp->line[ty])+(tx<<1);

  REPEAT8(

    // Get the next eight palette indexes
    pi=*((unsigned long *)b); b+=4;
    asm(
      "\n  movw %w0,%%fs\n"
      ASM_CORE_16("0") ASM_CORE_16("1")
      ASM_CORE_16("2") ASM_CORE_16("3")
      ASM_CORE_16("4") ASM_CORE_16("5")
      ASM_CORE_16("6") ASM_CORE_16("7")
      :
      : "rm" (bmp->seg) , "b" (pi) , "r" (pal), "r" (dest)
      : "%ecx","%ebx","%cc"
    );
    dest+=add_to_get_to_next_line;
  )

  return 0;
}

#define ASM_CORE_16_TRANS(PIX) \
"  movb %%bl,%%cl\n"\
"  andl $15,%%ecx\n"\
"  testb $15,%%ecx\n"\
"  jz trans16_" PIX "\n"\
"  movw (%2,%%ecx,4),%%cx\n"\
"  movw %%cx,%%fs:(%3)\n"\
"  trans16_" PIX ":\n"\
"  addl $2,%3\n"\
"  rorl $4,%%ebx\n\n"

static int draw_tile_16bpp_trans(BITMAP *bmp,int tx,int ty,
  unsigned char *b,int *pal)
{
  int y;
  unsigned long pi;
  unsigned short *dest;

  for (y=ty;y<(ty+8);y++)
  {
    dest=((unsigned short *)(bmp->line[y]))+tx;

    // Get the next eight palette indexes
    pi=*((unsigned long *)b); b+=4;
    if (pi==0) continue;
    asm(
      "\n  movw %w0,%%fs\n"

      ASM_CORE_16_TRANS("0") ASM_CORE_16_TRANS("1")
      ASM_CORE_16_TRANS("2") ASM_CORE_16_TRANS("3")
      ASM_CORE_16_TRANS("4") ASM_CORE_16_TRANS("5")
      ASM_CORE_16_TRANS("6") ASM_CORE_16_TRANS("7")
      :
      : "rm" (bmp->seg) , "b" (pi) , "r" (pal), "r" (dest)
      : "%ecx","%ebx","%cc"
    );
  }
  return 0;
}

static inline int draw_a_tile(BITMAP *bmp,int tx,int ty,int tile,int page)
{
  int i=0;
  unsigned char *b;
  int pal_no;
  int depth;

  pal_no=((tile&0xe000)>>13);
  pal_no+= ((int []){0,0x20,0x28,0}) [page&3];

  i = (tile & 0x03ff) + ((tile & 0x1000) >> 2);
  if ((tile&0x0c00)==0x0c00)
    i+=((tiledecoder2&0xf000)>>1);

  else if (tile&0x0400)
     i+=((tiledecoder1&0xf000)>>1);

  else if (tile&0x0800)
     i+=((tiledecoder2&0x0f00)<<3);

  i*=0x20;

  i&=0xfffff; // no seg faults please...!

  b=tile27+i;
  pal_no<<=4; // now a color offset

  depth=bitmap_color_depth(bmp);
  if (depth<=8)
  {
    draw_tile_8bpp_trans (bmp,tx,ty,b,our_color+pal_no);

   // Problem - shadows don't work if page 2 is opaque?
   // (yes)
  }
  else if (depth<=16)
  {
    if (page==2)
      draw_tile_16bpp      (bmp,tx,ty,b,our_color+pal_no);
    else
      draw_tile_16bpp_trans(bmp,tx,ty,b,our_color+pal_no);
   // Shadows seem to work okay here whatever...
  }
  else
    draw_tile_slow(bmp,tx,ty,b,our_color+pal_no);

  return 0;
}

static int draw_tiles(BITMAP *bmp,int page)
{
  int x,y;
  int offx=vpage[page].offx,offy=vpage[page].offy;

  // Run from 0 to ARCADE_H/8 inclusive which should be the same as
  // (bmp->h/8)-2 (because of 8 pixel messy border)

  for (y=0;y<= (bmp->h/8)-2 ;y++)
  {
    for (x=0;x<= (bmp->w/8)-2 ;x++)
    {
      int putx=(x<<3),puty=(y<<3);
      int tile=0;
      int getx,gety;

      getx=x+(offx>>3); gety=y+(offy>>3);
      getx&=63;         gety&=31;

      tile=MEMW(vidmem,0x000000+(page<<13)+(((gety<<6)+getx)<<1));

      // extra messy border so no clipping needed
      putx+=8;        puty+=8;
      putx-=(offx&7); puty-=(offy&7);

      draw_a_tile(bmp,putx,puty,tile,page);
    }
  }

  return 0;
}

struct tsprite
{
  unsigned char *b;
  int on;
  int x,y,w,h;
  int flipx,flipy;
  int pal_no;
};

inline void spr_core_slow
(BITMAP *bmp,unsigned long pi,int rx,int ry,int *our_palette)
{
  int our_col=pi&15,backcol=0;
  if (our_col)
  {
    if (our_col==15)
    {
      backcol=getpixel(bmp,rx,ry);
      if (backcol==0)
        our_col=our_palette[our_col];
      else
      {
        our_col=makecol(
          (getr(backcol)*3)>>2, 
          (getg(backcol)*3)>>2,
          (getb(backcol)*3)>>2);
      }
    }
    else
    {
      our_col=our_palette[our_col];
    }
    putpixel(bmp,rx,ry,our_col);
  }
}

#define SPR_CORE_ASM(PIX) \
    "  movb %b0,%%al\n"\
    "  andl $15,%%eax\n"\
    "  testb $15,%%al\n"\
    "  jz done_s_" PIX "\n"\
\
    "  cmpb $15,%%al\n"\
    "  jne no_shadow_" PIX "\n"\
\
    "  movb %%fs:(%1),%%bl\n"\
    "  cmpb $0,%%bl\n"\
    "  je no_shadow_" PIX "\n"\
\
    "  andw $0xda,%%bx\n"\
    "  rorw $1,%%bx\n"\
    "  movb %%bl,%%fs:(%1)\n"\
    "  jmp done_s_" PIX "\n"\
\
    "no_shadow_" PIX ":\n"\
\
    "  movw (%4,%%eax,4),%%bx\n"\
    "  movb %%bl,%%fs:(%1)\n"\
    "done_s_" PIX ":\n"\
    "  addl %2,%1\n"\
    "  rorl $4,%0\n"

// For 16-bit shadows
// (BELOW) Split rgb into %eax:00000gggggg00000 rrrrr000000bbbbb 
// Gives room to perform shifts and subtractions

#define SPR_CORE_ASM_16(PIX) \
    "  movb %b0,%%al\n"\
    "  andl $15,%%eax\n"\
    "  testb $15,%%al\n"\
    "  jz done_s16_" PIX "\n"\
\
    "  cmpb $15,%%al\n"\
    "  jne no_shadow16_" PIX "\n"\
\
    "  movw %%fs:(%1),%%bx\n"\
    "  cmpw $0,%%bx\n"\
    "  je no_shadow16_"PIX"\n"\
\
    "  movw %%bx,%%ax\n"\
    "  andw $0x07e0,%%ax\n"\
    "  andw $0xf81f,%%bx\n"\
    "  roll $16,%%eax\n"\
    "  movw %%bx,%%ax\n"\
    "#  andl $0x07e0f81f,%%eax\n"\
\
    "  movl %%eax,%%ebx\n"\
    "  sall $3,%%eax\n"\
    "  subl %%ebx,%%eax\n"\
    "  sarl $3,%%eax\n"\
\
    "  andl $0x07e0f81f,%%eax\n"\
    "  movw %%ax,%%bx\n"\
    "  rorl $16,%%eax\n"\
    "  orw  %%ax,%%bx\n"\
\
    "  movw %%bx,%%fs:(%1)\n"\
    "  jmp done_s16_" PIX "\n"\
\
    "no_shadow16_" PIX ":\n"\
\
    "  movw (%4,%%eax,4),%%bx\n"\
    "  movw %%bx,%%fs:(%1)\n"\
    "done_s16_" PIX ":\n"\
    "  addl %2,%1\n"\
    "  addl %2,%1\n"\
    "  rorl $4,%0\n"

// Probably need a seperate one for 15-bit
// For 15-bit shadows
// (BELOW) Split rgb into %eax:000000ggggg00000 0rrrrr00000bbbbb 
// Gives room to perform shifts and subtractions

#define SPR_CORE_ASM_15(PIX) \
    "  movb %b0,%%al\n"\
    "  andl $15,%%eax\n"\
    "  testb $15,%%al\n"\
    "  jz done_s15_" PIX "\n"\
\
    "  cmpb $15,%%al\n"\
    "  jne no_shadow15_" PIX "\n"\
\
    "  movw %%fs:(%1),%%bx\n"\
    "  cmpw $0,%%bx\n"\
    "  je no_shadow15_"PIX"\n"\
\
    "  movw %%bx,%%ax\n"\
    "  andw $0x03e0,%%ax\n"\
    "  andw $0x7c1f,%%bx\n"\
    "  roll $16,%%eax\n"\
    "  movw %%bx,%%ax\n"\
    "#  andl $0x03e07c1f,%%eax\n"\
\
    "  movl %%eax,%%ebx\n"\
    "  sall $3,%%eax\n"\
    "  subl %%ebx,%%eax\n"\
    "  sarl $3,%%eax\n"\
\
    "  andl $0x03e07c1f,%%eax\n"\
    "  movw %%ax,%%bx\n"\
    "  rorl $16,%%eax\n"\
    "  orw  %%ax,%%bx\n"\
\
    "  movw %%bx,%%fs:(%1)\n"\
    "  jmp done_s15_" PIX "\n"\
\
    "no_shadow15_" PIX ":\n"\
\
    "  movw (%4,%%eax,4),%%bx\n"\
    "  movw %%bx,%%fs:(%1)\n"\
    "done_s15_" PIX ":\n"\
    "  addl %2,%1\n"\
    "  addl %2,%1\n"\
    "  rorl $4,%0\n"

int draw_tsprite_medium(BITMAP *bmp,struct tsprite s)
{
  // Slow but sure way of drawing the sprites
  int osx=0,osy=0,sx=0,sy=0,rx=0,ry=0,dx=0,dy=0,lx=0,ly=0;
  // rx and ry is the rendering coord (different if flipped)
  // sx and sy is where rx and ry start from
  // dx and dy is the value to add to rx,ry each time
  // lx and ly are the limits of rx and ry
  // (if rx=lx, stop!)

  unsigned char *btm=s.b; // points to start of each line
  unsigned char *l=btm;   // goes along each pixel
  unsigned long pi=0;
  if (!s.on) return 1;

  dx=1; sx=s.x; lx=s.x+s.w; if (s.flipx) { dx=-1; sx=s.x+s.w-1; lx=s.x-1; }
  dy=1; sy=s.y; ly=s.y+s.h; if (s.flipy) { dy=-1; sy=s.y+s.h-1; ly=s.y-1; }
  btm=s.b;
  osx=sx; osy=sy;

  // Now do some clipping - take advantage of messy border horizontally
  // This is complex - basically we want it so
  // A) We draw 8 pixels at a time (using ALL of horizontal messy border)
  // but B) WE DON'T GO OVER THE EDGES OF THE BITMAP and
  // C) If we adjust a start position (sx or sy) we add the relevant
  //    offset to btm (our bitmap source)

  if (sx<8)
  {
    sx=(sx&7);
    if (!s.flipx)  btm+=(sx-osx)>>1;
    else return 0; //won't be seen
  }
  else if (sx>=ARCADE_W+8)
  {
    sx=(sx&7)+ARCADE_W+8;
    if (s.flipx) btm+=(osx-sx)>>1;
    else return 0; //won't be seen
  }

  if (!s.flipx)
  {
         if (lx<0) return 0; // won't be seen
    else if (lx>=ARCADE_W+8) lx=(lx&7)+ARCADE_W+8;
  }
  else
  {
         if (lx<0) lx=(lx&7);
    else if (lx>=ARCADE_W+8) return 0; // won't be seen
  }

  if (sy<8)
  {
    sy=8;
    if (!s.flipy) btm+=(s.w>>1)*(sy-osy);
  }
  if (sy>=ARCADE_H+8)
  {
    sy=ARCADE_H+8;
    if (s.flipy) btm+=(s.w>>1)*(osy-sy);
  }

  if (ly<8) ly=8;
  if (ly>=ARCADE_H+8) ly=ARCADE_H+8;

  // Okay - now if I've done that right, we can just blindly
  // render without worrying about seg faults...
  // Do I trust myself... yes!
/*
  if (dy>0) { if (ly<sy) return 0; }
  else      { if (ly>sy) return 0; }
  if (dx>0) { if (lx<sx) return 0; }
  else      { if (lx>sx) return 0; }
*/
  // Three times as much code (split the depth on the outside)

  if (bitmap_color_depth(bmp)<=8)
  {
    for   (ry=sy; ry!=ly; ry+=dy)
    {
      l=btm;
      for (rx=sx; rx!=lx; )
      {
        // Get the next eight palette indexes
        pi=*((unsigned long *)l); l+=4;
        asm (
          "  movw %w3,%%fs\n"
          SPR_CORE_ASM("0")
          SPR_CORE_ASM("1")
          SPR_CORE_ASM("2")
          SPR_CORE_ASM("3")
          SPR_CORE_ASM("4")
          SPR_CORE_ASM("5")
          SPR_CORE_ASM("6")
          SPR_CORE_ASM("7")
          :
          : "qm" (pi), "r" (bmp->line[ry]+rx), "qm" (dx), "rm" (bmp->seg),
            "qm" (our_color+(s.pal_no<<4))
          : "%eax", "%ebx", "%cc", "memory"
        );
        rx+=(dx*8);
      }
      btm+=s.w>>1;
    }
  }
  else if (bitmap_color_depth(bmp)<=15)
  {
    for   (ry=sy; ry!=ly; ry+=dy)
    {
      l=btm;
      for (rx=sx; rx!=lx; )
      {
        // Get the next eight palette indexes
        pi=*((unsigned long *)l); l+=4;
        asm (
          "  movw %w3,%%fs\n"
          SPR_CORE_ASM_15("0")
          SPR_CORE_ASM_15("1")
          SPR_CORE_ASM_15("2")
          SPR_CORE_ASM_15("3")
          SPR_CORE_ASM_15("4")
          SPR_CORE_ASM_15("5")
          SPR_CORE_ASM_15("6")
          SPR_CORE_ASM_15("7")
          :
          : "qm" (pi), "r" (bmp->line[ry]+rx+rx), "qm" (dx), "rm" (bmp->seg),
            "qm" (our_color+(s.pal_no<<4))
          : "%eax", "%cc", "%ebx","memory"
        );
        rx+=(dx*8);
      }
      btm+=s.w>>1;
    }
  }
  else if (bitmap_color_depth(bmp)<=16)
  {
    for   (ry=sy; ry!=ly; ry+=dy)
    {
      l=btm;
      for (rx=sx; rx!=lx; )
      {
        // Get the next eight palette indexes
        pi=*((unsigned long *)l); l+=4;
        asm (
          "  movw %w3,%%fs\n"
          SPR_CORE_ASM_16("0")
          SPR_CORE_ASM_16("1")
          SPR_CORE_ASM_16("2")
          SPR_CORE_ASM_16("3")
          SPR_CORE_ASM_16("4")
          SPR_CORE_ASM_16("5")
          SPR_CORE_ASM_16("6")
          SPR_CORE_ASM_16("7")
          :
          : "qm" (pi), "r" (bmp->line[ry]+rx+rx), "qm" (dx), "rm" (bmp->seg),
            "qm" (our_color+(s.pal_no<<4))
          : "%eax", "%cc", "%ebx","memory"
        );
        rx+=(dx*8);
      }
      btm+=s.w>>1;
    }
  }
  else
  {
    for   (ry=sy; ry!=ly; ry+=dy)
    {
      l=btm;
      for (rx=sx; rx!=lx; )
      {
        // Get the next eight palette indexes
        pi=*((unsigned long *)l); l+=4;
        REPEAT8(
          spr_core_slow(bmp,pi,rx,ry,our_color+(s.pal_no<<4)); pi>>=4; rx+=dx;
        );
      }
      btm+=s.w>>1;
    }
  }

  return 0;
}

static struct tsprite work_out_sprite_number(int i)
{
  // Decodes a sprite and returns a sprite structure
  static int size_table[16]=
  {16,16,32,16,
   16,32,32,32,
   64,32,32,64,
   64,64,128,128};
  int sproff=0x000000;
  struct tsprite spr={0};
  int c=0,offset=0;

  sproff+=i*8;

  spr.on=1;
  if ((MEMB(sprmem,sproff+0)&0x80)==0) spr.on=0;
  if (!spr.on) return spr;

  spr.x=((MEMB(sprmem,sproff+6)&1)<<8)+MEMB(sprmem,sproff+7);
  spr.y=((MEMB(sprmem,sproff+4)&1)<<8)+MEMB(sprmem,sproff+5);
  spr.x-=102; // align
  spr.y=256-spr.y;
  spr.y-=16; // align
  spr.x+=8; spr.y+=8; // Because of messy border

  spr.flipx=(MEMB(sprmem,sproff+6)&2)?1:0;
  spr.flipy=(MEMB(sprmem,sproff+4)&2)?1:0;

  // c is the code for the sprite size
  c=(MEMB(sprmem,sproff+1)>>4)&0xe; // (c=0,2,4... etc)
  spr.w=size_table[c+0];
  spr.h=size_table[c+1];

  offset=(MEMB(sprmem,sproff+1)&0x1f);
  offset<<=15;
  offset|=(MEMB(sprmem,sproff+2)<<7);

  // 32x64 hack from tmntemu and my hack for 16x32
  if (((offset&0x00700)==0x00200)&&(c==0x0a)) offset+=0x200;
  if (((offset&0x001e0)==0x00080)&&(c==0x04)) offset+=0x080;
  offset&=0xfffff; // no seg faults thanks!

  // bitmap is in sprite4/6 - which one?
  if (MEMB(sprmem,sproff+3)&0x10) spr.b=sprite6+offset;
  else                     spr.b=sprite4+offset;

  // Now get the palette number
  spr.pal_no=((MEMB(sprmem,sproff+3)&0x0f)+0x10);

  return spr;
}

#define FRAME_MAX (64)
static struct tsprite frame_tspr[FRAME_MAX][64]={{{0}}};
static int frame_capture=0;
static int clean_the_slate=1; // next time you work_out_sprites

int work_out_sprites()
{
  // Work out the sprites at a given instance
  // And stash in frame_tspr
  int i;

  if (clean_the_slate) frame_capture=0;
  clean_the_slate=0;

  for (i=0;i<64;i++)
    frame_tspr[frame_capture][i]=work_out_sprite_number(i);

  if (frame_capture<(FRAME_MAX-1))
    frame_capture++;
  return 0;
}

int we_do_sprites_behind=0; // set by the machine loop (rom.c)

int trippy_sprites=0;

static int now_draw_the_sprites(BITMAP *bmp)
{
  int i=0,f=0;

  if (trippy_sprites==-1)
  {
    for (i=63;i>=0;i--)
      for (f=0;f<frame_capture;f++)
        draw_tsprite_medium(bmp,frame_tspr[f][i]);
  }
  else
  {
    int frame_to_draw=0;
    // Draw the most recent frame
    frame_to_draw=0;
    if (frame_capture>0) frame_to_draw=frame_capture-1;

    for (i=63;i>=0;i--)
      draw_tsprite_medium(bmp,frame_tspr[frame_to_draw][i]);
  }

  return 0;
}

int draw_graphics(BITMAP *bmp)
{
  int depth=0;

    SBK
  update_vpage(0);
    SBK
  update_vpage(1);
    SBK
  update_vpage(2);
    SBK

  // tile routines need tiledecoders defined
  tiledecoder1 = MEMW(vidmem,0x006b00);
  tiledecoder2 = MEMW(vidmem,0x006e00);
  // convert all the palettes to our color codes (our_color[])
    SBK
  depth=bitmap_color_depth(bmp);
  convert_all_to_our_colors(depth);
    SBK

  clear(bmp);

  if (do_int5)
  {
    SBK
    if (layer&4) draw_tiles(bmp,2);
    SBK
    if (layer&2) draw_tiles(bmp,1);
    SBK

    if (we_do_sprites_behind)
    {
    SBK
      if (layer&8) now_draw_the_sprites(bmp);
    SBK
      if (layer&2) draw_tiles(bmp,1);
    SBK
    }
    else
    {
    SBK
      if (layer&2) draw_tiles(bmp,1);
    SBK
      if (layer&8) now_draw_the_sprites(bmp);
    SBK
    }
  }

    SBK
  if (layer&1) draw_tiles(bmp,0);
    SBK

  clean_the_slate=1;

  return 0;
}

